#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import json

import netius.clients


class PushiChannel(netius.observer.Observable):
    def __init__(self, owner, name):
        netius.observer.Observable.__init__(self)
        self.owner = owner
        self.name = name
        self.data = None
        self.subscribed = False

    def set_subscribe(self, data):
        alias = data["alias"] if data else []
        for name in alias:
            channel = PushiChannel(self.owner, name)
            self.owner.channels[name] = channel
            self.owner.on_subscribe_pushi(name, {})

        self.data = data
        self.subscribed = True
        self.trigger("subscribe", self, data)

    def set_unsubscribe(self, data):
        alias = data["alias"] if data else []
        for name in alias:
            self.owner.on_subscribe_pushi(name, {})

        self.subscribed = False
        self.trigger("unsubscribe", self, data)

    def set_latest(self, data):
        self.trigger("latest", self, data)

    def set_message(self, event, data, mid=None, timestamp=None):
        self.trigger(event, self, data, mid=mid, timestamp=timestamp)

    def send(self, event, data, persist=True):
        self.owner.send_channel(event, data, self.name, persist=persist)

    def unsubscribe(self, callback=None):
        self.owner.unsubscribe_pushi(self.name, callback=callback)

    def latest(self, skip=0, count=10, callback=None):
        self.owner.latest_pushi(self.name, skip=skip, count=count, callback=callback)


class PushiProtocol(netius.clients.WSProtocol):
    PUXIAPP_URL = "wss://puxiapp.com/"
    """ The default PuxiApp URL that is going to be used
    to establish new client's connections """

    def __init__(self, *args, **kwargs):
        netius.clients.WSProtocol.__init__(self, *args, **kwargs)
        self.base_url = None
        self.client_key = None
        self.api = None
        self.url = None
        self.state = "disconnected"
        self.socket_id = None
        self.channels = dict()

    def connect_pushi(
        self, url=None, client_key=None, api=None, callback=None, loop=None
    ):
        cls = self.__class__

        self.base_url = url or cls.PUXIAPP_URL
        self.client_key = client_key
        self.api = api
        self.url = self.base_url + self.client_key

        if callback:
            self.bind("connect_pushi", callback, oneshot=True)

        return self.connect_ws(self.url, loop=loop)

    def receive_ws(self, data):
        data = data.decode("utf-8")
        data_j = json.loads(data)

        is_connected = (
            self.state == "disconnected"
            and data_j["event"] == "pusher:connection_established"
        )

        if is_connected:
            data = json.loads(data_j["data"])
            self.on_connect_pushi(data)
        elif self.state == "connected":
            self.on_message_pushi(data_j)

    def on_connect_pushi(self, data):
        self.socket_id = data["socket_id"]
        self.state = "connected"
        self.trigger("connect_pushi", self)

    def on_disconnect_pushi(self, data):
        self.socket_id = None
        self.channels = dict()
        self.state = "disconnected"
        self.trigger("disconnect_pushi", self)

    def on_message_pushi(self, data_j):
        # unpacks the complete set of information from the JSON based
        # data structure so that it gets processed in the method
        data = data_j["data"]
        event = data_j["event"]
        channel = data_j["channel"]
        mid = data_j.get("mid", None)
        timestamp = data_j.get("timestamp", None)

        # tries to gather the channel object/information reference
        # for the channel that received the message and verifies if
        # the current channel is peer related or not
        _channel = self.channels.get(channel, None)
        is_peer = channel.startswith("peer-")

        # in case no channel information is found for the channel
        # (no subscription) and the channel is not peer related the
        # message is ignored as there's no channel subscription
        if channel and not _channel and not is_peer:
            return

        if event == "pusher_internal:subscription_succeeded":
            data = json.loads(data_j["data"])
            self.on_subscribe_pushi(channel, data)

        elif event == "pusher_internal:unsubscription_succeeded":
            data = json.loads(data_j["data"])
            self.on_unsubscribe_pushi(channel, data)

        elif event == "pusher_internal:latest":
            data = json.loads(data_j["data"])
            self.on_latest_pushi(channel, data)

        elif event == "pusher:member_added":
            member = json.loads(data_j["member"])
            self.on_member_added_pushi(channel, member)

        elif event == "pusher:member_removed":
            member = json.loads(data_j["member"])
            self.on_member_removed_pushi(channel, member)

        self.trigger(event, self, data, channel, mid=mid, timestamp=timestamp)
        if _channel:
            _channel.set_message(event, data, mid=mid, timestamp=timestamp)

    def on_subscribe_pushi(self, channel, data):
        _channel = self.channels[channel]
        _channel.set_subscribe(data)
        self.trigger("subscribe", self, channel, data)

    def on_unsubscribe_pushi(self, channel, data):
        _channel = self.channels[channel]
        del self.channels[channel]
        _channel.set_unsubscribe(data)
        self.trigger("unsubscribe", self, channel, data)

    def on_latest_pushi(self, channel, data):
        _channel = self.channels[channel]
        _channel.set_latest(data)
        self.trigger("latest", self, channel, data)

    def on_member_added_pushi(self, channel, member):
        pass

    def on_member_removed_pushi(self, channel, member):
        pass

    def subscribe_pushi(self, channel, channel_data=None, force=False, callback=None):
        exists = channel in self.channels
        if exists and not force:
            return

        is_private = self._is_private(channel)
        if is_private:
            self._subscribe_private(channel, channel_data=channel_data)
        else:
            self._subscribe_public(channel)

        name = channel
        channel = PushiChannel(self, name)
        self.channels[name] = channel

        if callback:
            channel.bind("subscribe", callback, oneshot=True)

        return channel

    def unsubscribe_pushi(self, channel, callback=None):
        exists = channel in self.channels
        if not exists:
            return

        self._unsubscribe(channel)

        name = channel
        channel = self.channels[name]

        if callback:
            channel.bind("unsubscribe", callback, oneshot=True)

        return channel

    def latest_pushi(self, channel, skip=0, count=10, callback=None):
        exists = channel in self.channels or channel.startswith("peer-")
        if not exists:
            return

        self._latest(channel, skip=skip, count=count)

        name = channel
        channel = self._ensure_channel(name)

        if callback:
            channel.bind("latest", callback, oneshot=True)

        return channel

    def send_event(self, event, data, echo=False, persist=True, callback=None):
        json_d = dict(event=event, data=data, echo=echo, persist=persist)
        self.send_pushi(json_d, callback=callback)

    def send_channel(
        self, event, data, channel, echo=False, persist=True, callback=None
    ):
        json_d = dict(
            event=event, data=data, channel=channel, echo=echo, persist=persist
        )
        self.send_pushi(json_d, callback=callback)

    def send_pushi(self, json_d, callback=None):
        data = json.dumps(json_d)
        self.send_ws(data, callback=callback)

    def _ensure_channel(self, name):
        if name in self.channels:
            return self.channels[name]
        channel = PushiChannel(self, name)
        self.channels[name] = channel
        return channel

    def _subscribe_public(self, channel):
        self.send_event("pusher:subscribe", dict(channel=channel))

    def _subscribe_private(self, channel, channel_data=None):
        if not self.api:
            raise RuntimeError("No private app available")
        auth = self.api.authenticate(channel, self.socket_id)
        self.send_event(
            "pusher:subscribe",
            dict(channel=channel, auth=auth, channel_data=channel_data),
        )

    def _unsubscribe(self, channel):
        self.send_event("pusher:unsubscribe", dict(channel=channel))

    def _latest(self, channel, skip=0, count=10):
        self.send_event("pusher:latest", dict(channel=channel, skip=skip, count=count))

    def _is_private(self, channel):
        return (
            channel.startswith("private-")
            or channel.startswith("presence-")
            or channel.startswith("personal-")
        )


class PushiClient(netius.clients.WSClient):
    protocol = PushiProtocol

    @classmethod
    def connect_pushi_s(
        cls, url=None, client_key=None, api=None, callback=None, loop=None
    ):
        protocol = cls.protocol()
        return protocol.connect_pushi(
            url=url, client_key=client_key, api=api, callback=callback, loop=loop
        )


if __name__ == "__main__":

    def on_message(channel, data, mid=None, timestamp=None):
        print("Received %s" % data)
        channel.unsubscribe(callback=on_unsubscribe)

    def on_unsubscribe(channel, data):
        connection = channel.owner
        client = connection.owner
        client.close()

    def on_latest(channel, data):
        name = data["name"]
        events = data["events"]
        print("Received the latest %d event(s) for channel %s" % (len(events), name))

    def on_subscribe(channel, data):
        channel.send("message", "Hello World", persist=False)
        channel.bind("message", on_message)
        channel.latest(count=20, callback=on_latest)

    def on_connect(protocol):
        protocol.subscribe_pushi("global", callback=on_subscribe)

    def register_timer(protocol):
        def timer():
            print("Waiting for events(s) on channel global ...")
            protocol.delay(timer, timeout=5)

        protocol.delay(timer, timeout=5)

    url = netius.conf("PUSHI_URL")
    client_key = netius.conf("PUSHI_KEY")

    loop, protocol = PushiClient.connect_pushi_s(
        url=url, client_key=client_key, callback=on_connect
    )

    register_timer(protocol)

    loop.run_forever()
    loop.close()
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier
import appier_extras

import pushi


class PushiApp(appier.APIApp):
    def __init__(self, state=None, *args, **kwargs):
        appier.APIApp.__init__(
            self, name="pushi", parts=(appier_extras.AdminPart,), *args, **kwargs
        )
        self.state = state

    def info_dict(self):
        info = appier.APIApp.info_dict(self)
        if not self.state:
            return info
        server = self.state.server
        info["service"] = server.info_dict()
        return info

    def auth(self, app_id, app_key, app_secret, **kwargs):
        app = pushi.App.get(ident=app_id, key=app_key, secret=app_secret)
        if not app:
            raise RuntimeError("Invalid credentials provided")

    def on_login(self, sid, secret, app_id, app_key, app_secret, **kwargs):
        appier.APIApp.on_login(self, sid, secret, **kwargs)
        self.session["app_id"] = app_id
        self.session["app_key"] = app_key
        self.session["app_secret"] = app_secret

    def on_logout(self):
        appier.APIApp.on_logout(self)
        if not self.session:
            return
        if "app_id" in self.session:
            del self.session["app_id"]
        if "app_key" in self.session:
            del self.session["app_key"]
        if "app_secret" in self.session:
            del self.session["app_secret"]

    def _version(self):
        return "0.3.1"

    def _description(self):
        return "Pushi"

    def _observations(self):
        return (
            "Simple yet powerful infra-structure for handling of WebSocket connections"
        )


if __name__ == "__main__":
    app = PushiApp()
    app.serve()
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier
import appier_extras


class PushiBase(appier_extras.admin.Base):
    instance = appier.field(index=True, safe=True, immutable=True)

    @classmethod
    def get(cls, *args, **kwargs):
        request = appier.get_request()
        app_id = request.session.get("app_id", None)
        if app_id:
            kwargs["instance"] = app_id
        return super(PushiBase, cls).get(cls, *args, **kwargs)

    @classmethod
    def find(cls, *args, **kwargs):
        request = appier.get_request()
        app_id = request.session.get("app_id", None)
        if app_id:
            kwargs["instance"] = app_id
        return super(PushiBase, cls).find(cls, *args, **kwargs)

    @classmethod
    def count(cls, *args, **kwargs):
        request = appier.get_request()
        app_id = request.session.get("app_id", None)
        if app_id:
            kwargs["instance"] = app_id
        return super(PushiBase, cls).count(cls, *args, **kwargs)

    @classmethod
    def exists(cls, *args, **kwargs):
        previous = cls.find(*args, **kwargs)
        return previous[0] if previous else None

    def pre_create(self):
        appier_extras.admin.Base.pre_create(self)
        if self.app_id:
            self.instance = self.app_id

    @property
    def state(self):
        app = appier.get_app()
        return app.state

    @property
    def app_id(self):
        if hasattr(self, "instance"):
            return self.instance
        request = appier.get_request()
        return request.session.get("app_id", None)

    @property
    def app_key(self):
        if not self.app_id:
            return None
        return self.state.app_id_to_app_key(self.app_id)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier


class SocketController(appier.Controller):
    @appier.private
    @appier.route("/sockets", "GET")
    def list(self):
        app_id = self.session.get("app_id", None)
        state = self.state.get_state(app_id=app_id)

        sockets = []

        for socket_id, channel in state.socket_channels.iteritems():
            socket = dict(socket_id=socket_id, channel=channel)
            sockets.append(socket)

        return dict(sockets=sockets)

    @appier.private
    @appier.route("/sockets/<socket_id>", "POST")
    def show(self, socket_id):
        app_id = self.session.get("app_id", None)
        state = self.state.get_state(app_id=app_id)
        channels = state.socket_channels.get(socket_id, [])

        return dict(channels=channels)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier

from . import base


class APN(base.PushiBase):
    token = appier.field(index=True)

    event = appier.field(index=True)

    @classmethod
    def validate(cls):
        return super(APN, cls).validate() + [
            appier.not_null("token"),
            appier.not_empty("token"),
            appier.not_null("event"),
            appier.not_empty("event"),
        ]

    @classmethod
    def list_names(cls):
        return ["token", "event"]

    @classmethod
    def _underscore(cls, plural=True):
        return "apns" if plural else "apn"

    @classmethod
    def _readable(cls, plural=False):
        return "APNs" if plural else "APN"

    def pre_update(self):
        base.PushiBase.pre_update(self)
        previous = self.__class__.get(id=self.id)
        self.state and self.state.apn_handler.remove(
            previous.app_id, previous.token, previous.event
        )

    def post_create(self):
        base.PushiBase.pre_create(self)
        self.state and self.state.apn_handler.add(self.app_id, self.token, self.event)

    def post_update(self):
        base.PushiBase.post_update(self)
        self.state and self.state.apn_handler.add(self.app_id, self.token, self.event)

    def post_delete(self):
        base.PushiBase.post_delete(self)
        self.state and self.state.apn_handler.remove(
            self.app_id, self.token, self.event
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import json

import netius.clients

import pushi

from . import handler


class WebHandler(handler.Handler):
    """
    Event handler to be used for web based "hooks".
    This handler provides the abstraction for the HTTP
    client based callbacks.
    """

    def __init__(self, owner):
        handler.Handler.__init__(self, owner, name="web")
        self.subs = {}

    def send(self, app_id, event, json_d, invalid={}):
        # retrieves the reference to the app structure associated with the
        # id for which the message is being send
        app = self.owner.get_app(app_id=app_id)

        # retrieves the app key for the retrieved app by unpacking the current
        # app structure into the appropriate values
        app_key = app.key

        # saves the original event name for the received event, so that it may
        # be used latter for debugging/log purposes
        root_event = event

        # resolves the complete set of (extra) channels for the provided
        # event assuming that it may be associated with alias, then creates
        # the complete list of event containing also the "extra" events
        extra = self.owner.get_channels(app_key, event)
        events = [event] + extra

        # retrieves the complete set of subscriptions for the current web
        # infra-structure to be able to resolve the appropriate urls
        subs = self.subs.get(app_id, {})

        # creates the initial list of URLs to be notified and then populates
        # the list with the various URL associated with the complete set of
        # resolved events, note that a set is created at the end so that one
        # URL gets notified only once (no double notifications)
        urls = []
        for event in events:
            _urls = subs.get(event, [])
            urls.extend(_urls)
        urls = set(urls)
        count = len(urls)

        # prints a logging message about the various (web) subscriptions
        # that were found for the event that was triggered
        self.logger.debug("Found %d web subscription(s) for '%s'" % (count, root_event))

        # serializes the JSON message so that it's possible to send it using
        # the HTTP client to the endpoints and then creates the map of headers
        # that is going to be used in the post messages to be sent
        data = json.dumps(json_d)
        headers = {"content-type": "application/json"}

        # creates the on message function that is going to be used at the end of
        # the request to be able to close the protocol, this is a clojure and so
        # current local variables will be exposed to the method
        def on_message(protocol, parser, message):
            protocol.close()

        # creates the on close function that will be responsible for the stopping
        # of the loop as defined by the web implementation
        def on_finish(protocol):
            netius.compat_loop(loop).stop()

        # iterates over the complete set of URLs that are going to
        # be notified about the message, each of them is going to
        # received an HTTP post request with the data
        for url in urls:
            # in case the current token is present in the current
            # map of invalid items must skip iteration as the message
            # has probably already been sent "to the target URL"
            if url in invalid:
                continue

            # prints a debug message about the web message that
            # is going to be sent (includes URL)
            self.logger.debug("Sending POST request to '%s'" % url)

            # creates the HTTP protocol to be used in the POST request and
            # sets the headers and the data then registers for the message
            # event so that the loop and protocol may be closed
            loop, protocol = netius.clients.HTTPClient.post_s(
                url, headers=headers, data=data
            )
            protocol.bind("message", on_message)
            protocol.bind("finish", on_finish)
            loop.run_forever()

            # adds the current URL to the list of invalid items for
            # the current message sending stream
            invalid[url] = True

    def load(self):
        subs = pushi.Web.find()
        for sub in subs:
            app_id = sub.app_id
            url = sub.url
            event = sub.event
            self.add(app_id, url, event)

    def add(self, app_id, url, event):
        events = self.subs.get(app_id, {})
        urls = events.get(event, [])
        urls.append(url)
        events[event] = urls
        self.subs[app_id] = events

    def remove(self, app_id, url, event):
        events = self.subs.get(app_id, {})
        urls = events.get(event, [])
        if url in urls:
            urls.remove(url)

    def subscriptions(self, url=None, event=None):
        filter = dict()
        if url:
            filter["url"] = url
        if event:
            filter["event"] = event
        subscriptions = pushi.Web.find(map=True, **filter)
        return dict(subscriptions=subscriptions)

    def subscribe(self, web, auth=None, unsubscribe=True):
        self.logger.debug("Subscribing '%s' for '%s'" % (web.url, web.event))

        is_private = (
            web.event.startswith("private-")
            or web.event.startswith("presence-")
            or web.event.startswith("peer-")
            or web.event.startswith("personal-")
        )

        is_private and self.owner.verify(web.app_key, web.url, web.event, auth)
        unsubscribe and self.unsubscribe(web.url, force=False)

        exists = pushi.Web.exists(url=web.url, event=web.event)
        if exists:
            web = exists
        else:
            web.save()

        self.logger.debug("Subscribed '%s' for '%s'" % (web.url, web.event))

        return web

    def unsubscribe(self, url, event=None, force=True):
        self.logger.debug("Unsubscribing '%s' from '%s'" % (url, event or "*"))

        kwargs = dict(url=url, raise_e=force)
        if event:
            kwargs["event"] = event

        web = pushi.Web.get(**kwargs)
        if not web:
            return None

        web.delete()

        self.logger.debug("Unsubscribed '%s' for '%s'" % (url, event or "*"))

        return web

    def unsubscribes(self, url, event=None):
        kwargs = dict(token=url)
        if event:
            kwargs["event"] = event

        webs = pushi.Web.find(**kwargs)
        for web in webs:
            web.delete()

        return webs



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import sys
import time
import json
import hmac
import uuid
import copy
import hashlib
import datetime
import threading

base_dir = os.path.normpath((os.path.dirname(__file__) or ".") + "/../..")
if not base_dir in sys.path:
    sys.path.insert(0, base_dir)

import pushi
import appier

from pushi.base import apn
from pushi.base import web


class AppState(object):
    """
    The state object that defined the various state variables
    for an app registered in the system. There should be one
    of this objects per each application loaded.
    """

    def __init__(self, app_id, app_key):
        self.app_id = app_id
        self.app_key = app_key
        self.alias = {}
        self.alias_i = {}
        self.socket_channels = {}
        self.channel_sockets = {}
        self.channel_info = {}
        self.channel_socket_data = {}


class State(appier.Mongo):
    """
    Main logic of the pushi infra-structure, this class
    should contain the main structures and operations that
    control a system for push notifications.

    It should run in an asynchronous nonblocking fashion to
    avoid the typical locking related problems (eg: dead locks)
    and at the same time handle the c10k problem.

    The structure of the system is based on the encapsulation
    of both the (async) server and the web based app that handles
    the HTTP based requests (REST API).
    """

    def __init__(self):
        appier.Mongo.__init__(self)
        self.app = None
        self.server = None
        self.apn_handler = None
        self.handlers = []
        self.app_id_state = {}
        self.app_key_state = {}

    def load(self, app, server):
        # sets the references to both the app and the server in the
        # current instance, this values are going to be used latter
        self.app = app
        self.server = server

        # "moves" the (in memory) logging handlers of the app to the
        # server so that they share a common logging infrastructure
        handlers = self.app.handlers
        level = self.app.level
        self.server.handlers = handlers
        self.server.level = level

        # registers for the various base events in the server so that
        # it's able to properly update the current state of the application
        # to the new state (according to the operation)
        self.server.bind("connect", self.connect)
        self.server.bind("disconnect", self.disconnect)
        self.server.bind("subscribe", self.subscribe)
        self.server.bind("unsubscribe", self.unsubscribe)
        self.server.bind("validate", self.validate)

        # retrieves the various environment variable values that are going
        # to be used in the starting of both the app server and the proper
        # pushi server (most of the values have default values)
        APP_SERVER = os.environ.get("APP_SERVER", "netius")
        APP_HOST = os.environ.get("APP_HOST", "127.0.0.1")
        APP_PORT = int(os.environ.get("APP_PORT", "8080"))
        APP_SSL = bool(int(os.environ.get("APP_SSL", "0")))
        APP_SSL_KEY = os.environ.get("APP_SSL_KEY", None)
        APP_SSL_CER = os.environ.get("APP_SSL_CER", None)
        SERVER_HOST = os.environ.get("SERVER_HOST", "127.0.0.1")
        SERVER_PORT = int(os.environ.get("SERVER_PORT", "9090"))
        SERVER_SSL = bool(int(os.environ.get("SERVER_SSL", "0")))
        SERVER_SSL_KEY = os.environ.get("SERVER_SSL_KEY", None)
        SERVER_SSL_CER = os.environ.get("SERVER_SSL_CER", None)

        # creates the named argument for both the app server and the proper
        # pushi server so that they are correctly initialized and bound to
        # the proper ports (infinite loop)
        app_kwargs = dict(
            server=APP_SERVER,
            host=APP_HOST,
            port=APP_PORT,
            ssl=APP_SSL,
            key_file=APP_SSL_KEY,
            cer_file=APP_SSL_CER,
        )
        server_kwargs = dict(
            host=SERVER_HOST,
            port=SERVER_PORT,
            ssl=SERVER_SSL,
            key_file=SERVER_SSL_KEY,
            cer_file=SERVER_SSL_CER,
        )

        # creates the threads that will be used as containers for the app and
        # the pushi server and then starts them with the proper arguments
        threading.Thread(target=self.app.serve, kwargs=app_kwargs).start()
        threading.Thread(target=self.server.serve, kwargs=server_kwargs).start()

        # starts the loading process of the various (extra handlers) that are
        # going to be used in the pushi infra-structure (eg: apn, gcm, etc.)
        self.load_handlers()

        # loads the various alias relations for the current infra-structure so
        # that the personal channels are able to correctly work
        self.load_alias()

    def load_handlers(self):
        self.apn_handler = apn.APNHandler(self)
        self.web_handler = web.WebHandler(self)

        self.apn_handler.load()
        self.web_handler.load()

        self.handlers.append(self.apn_handler)
        self.handlers.append(self.web_handler)

    def load_alias(self):
        """
        Loads the complete set of alias (channels that represent) the
        same for the current context, this may be used for a variety
        of reasons including the personal channels.
        """

        # retrieves the complete set of subscriptions from the currently
        # associated data source reference and then uses them to create
        # the complete personal to proper channel relation
        subs = pushi.Subscription.find()
        for sub in subs:
            app_id = sub.app_id
            user_id = sub.user_id
            event = sub.event
            app_key = self.app_id_to_app_key(app_id)
            self.add_alias(app_key, "personal-" + user_id, event)

    def add_alias(self, app_key, channel, alias):
        self.app.logger.debug(
            "Adding '%s' into '%s' for app key '%s'" % (alias, channel, app_key)
        )

        state = self.get_state(app_key=app_key)
        alias_l = state.alias.get(channel, [])
        if alias in alias_l:
            return

        alias_l.append(alias)
        state.alias[channel] = alias_l

        alias_l = state.alias_i.get(alias, [])
        if channel in alias_l:
            return

        alias_l.append(channel)
        state.alias_i[alias] = alias_l

    def remove_alias(self, app_key, channel, alias):
        self.app.logger.debug(
            "Removing '%s' from '%s' for app key '%s'" % (alias, channel, app_key)
        )

        state = self.get_state(app_key=app_key)
        alias_l = state.alias.get(channel, [])
        if not alias in alias_l:
            return

        alias_l.remove(alias)
        if not alias_l:
            del state.alias[channel]

        alias_l = state.alias_i.get(alias, [])
        if not channel in alias_l:
            return

        alias_l.remove(channel)
        if not alias_l:
            del state.alias_i[alias]

    def connect(self, connection, app_key, socket_id):
        pass

    def disconnect(self, connection, app_key, socket_id):
        # in case no app key or socket id is defined must return
        # immediately because it's not possible to perform the
        # disconnect operation in such conditions, possible a non
        # established connection is attempting to disconnect
        if not app_key:
            return
        if not socket_id:
            return

        # verifies if there's an app value available for the app key
        # in case there's not, returns immediately, as it's not possible
        # to disconnect a connection without an associated/valid app
        if not self.has_app(app_key=app_key):
            return

        # retrieves the current state of the app using the app key and
        # then uses it to retrieve the complete set of channels that the
        # socket is subscribed and then unsubscribe it from them then
        # removes the reference of the socket in the socket channels map
        state = self.get_state(app_key=app_key)
        channels = state.socket_channels.get(socket_id, [])
        channels = copy.copy(channels)
        for channel in channels:
            self.unsubscribe(connection, app_key, socket_id, channel)
        if socket_id in state.socket_channels:
            del state.socket_channels[socket_id]

    def subscribe(
        self,
        connection,
        app_key,
        socket_id,
        channel,
        auth=None,
        channel_data=None,
        force=False,
    ):
        # checks if the the channel to be registered is considered private
        # (either private, presence or peer) and in case it's private verifies
        # if the correct credentials (including auth token) are valid
        is_private = (
            channel.startswith("private-")
            or channel.startswith("presence-")
            or channel.startswith("peer-")
            or channel.startswith("personal-")
        )
        if is_private and not force:
            self.verify(app_key, socket_id, channel, auth)

        # verifies if the current channel is of type personal and in
        # case it's retrieves it's alias (channels) and subscribes to
        # all of them (as expected), then return immediately
        is_personal = channel.startswith("personal-")
        if is_personal:
            channels = self.get_alias(app_key, channel)
            for channel in channels:
                self.subscribe(
                    connection,
                    app_key,
                    socket_id,
                    channel,
                    auth=auth,
                    channel_data=channel_data,
                    force=True,
                )
            return

        # verifies if the channel is of type presence (prefix based
        # verification) and in case it's not invalidate the channel
        # data as channel data is not valid
        is_presence = channel.startswith("presence-")
        if not is_presence:
            channel_data = None

        # verifies if the current connection (by socket id) is already
        # registered to the channel and in case it's unsubscribes the
        # connection from it (avoids duplicated registration)
        is_subscribed = self.is_subscribed(app_key, socket_id, channel)
        if is_subscribed:
            self.unsubscribe(connection, app_key, socket_id, channel)

        # retrieves the global state structure for the provided API key
        # and also creates the tuple that encapsulates both the channel
        # and the socket id (unique identification)
        state = self.get_state(app_key=app_key)
        channel_socket = (channel, socket_id)

        # retrieves the complete set of channels for the socket id and
        # adds the current channel to it (subscription) then updates the
        # association between the socket id and the channels
        channels = state.socket_channels.get(socket_id, [])
        subscribed = channel in channels
        if subscribed:
            raise RuntimeError("Channel already subscribed")
        channels.append(channel)
        state.socket_channels[socket_id] = channels

        # retrieves the complete set of sockets for the channels (inverted)
        # association and adds the current socket id to the list then
        # re-updates the inverted map with the sockets list
        sockets = state.channel_sockets.get(channel, [])
        subscribed = socket_id in sockets
        if subscribed:
            raise RuntimeError("Socket already subscribed")
        sockets.append(socket_id)
        state.channel_sockets[channel] = sockets

        # in case there's no channel data to be used to change
        # metadata and in additional processing must return
        # immediately as there's nothing else remaining to be
        # done in the subscription process
        if not channel_data:
            return

        # unpacks the information from the channel data, defaulting
        # some of the values to their fallback values
        user_id = channel_data["user_id"]
        is_peer = channel_data.get("peer", False)

        # "saves" the channel data for the channel socket tuple in the
        # appropriate data structure to be used latter
        state.channel_socket_data[channel_socket] = channel_data

        # retrieves the channel info for the channel that is going to be
        # subscribed and "unpacks" the complete state information from it
        # so that it may be used and updated
        info = state.channel_info.get(channel, {})
        users = info.get("users", {})
        members = info.get("members", {})
        conns = info.get("conns", [])
        user_count = info.get("user_count", 0)

        # adds the current connection to the list of connection for the
        # the current channel (state information update)
        conns.append(connection)

        # retrieves the list of connection to the current user id that is
        # going to be used and adds the current connection, then re-updates
        # the user connections list and the channel data
        user_conns = users.get(user_id, [])
        user_conns.append(connection)
        users[user_id] = user_conns
        members[user_id] = channel_data

        # verifies if the current subscription is going to create a new user
        # subscription (that must be logger) this is the case if the number
        # of connection currently subscribed is one
        is_new = len(user_conns) == 1
        if is_new:
            user_count += 1

        # updates the info dictionary with the new complete set of values for
        # the channel information
        info["users"] = users
        info["members"] = members
        info["conns"] = conns
        info["user_count"] = user_count
        state.channel_info[channel] = info

        # subscribes all of the peer channels associated with the current
        # presence channel that is being subscribed, this may represent some
        # overhead but provides peer to peer communication
        is_peer and self.subscribe_peer_all(app_key, connection, channel)

        # in case the connection does not represent a new user logging in must
        # return immediately, because there's nothing remaining to be done
        if not is_new:
            return

        # creates the JSON dictionary containing the event to be sent to all
        # of the socket in the channel indicating the presence change for the
        # "new" user that has just logged in (member added)
        json_d = dict(
            event="pusher:member_added",
            member=json.dumps(channel_data),
            channel=channel,
        )

        # iterates over the complete set of connections currently subscribed
        # to the channel, in order to be notify them about the member added
        for _connection in conns:
            if _connection == connection:
                continue
            _connection.send_pushi(json_d)

            # in case the connection is not of type peer there's nothing
            # else to be done related with the other connections
            if not is_peer:
                continue

            # retrieves the socket id of the current connection in iteration
            # and uses it to construct the channel socket id tuple to try to
            # retrieve the channel data for the socket in case it does not
            # exists skips the current step, no need to subscribe to chat
            # specific channel (because there's no channel data)
            _socket_id = _connection.socket_id
            _channel_socket = (channel, _socket_id)
            _channel_data = state.channel_socket_data.get(_channel_socket)
            if not _channel_data:
                continue

            # retrieves the user id of the current channel data for the user
            # in iteration and the uses this value to subscribe it for the
            # peer channel with the current member (that was just added)
            _user_id = _channel_data["user_id"]
            self.subscribe_peer(app_key, _connection, channel, user_id, _user_id)

    def unsubscribe(self, connection, app_key, socket_id, channel):
        # checks if the current channel is a private one and in case
        # it's runs the unsubscription operation for all of the alias
        # channels associated with this personal one
        is_personal = channel.startswith("personal-")
        if is_personal:
            channels = self.get_alias(app_key, channel)
            for channel in channels:
                self.unsubscribe(connection, app_key, socket_id, channel)
            return

        # runs the validation process for the subscription to the
        # current channel, meaning that in case the current socket
        # is not subscribed to the channel an exception will be raised
        self.validate(connection, app_key, socket_id, channel)

        # uses the provided app key to retrieve the state of the
        # app and then creates the channel socket tuple that is
        # going to be used for unique identification
        state = self.get_state(app_key=app_key)
        channel_socket = (channel, socket_id)

        # retrieves the list of channels for which the provided socket
        # id is currently subscribed and removes the current channel
        # from that list in case it exists there
        channels = state.socket_channels.get(socket_id, [])
        if channel in channels:
            channels.remove(channel)

        # retrieves the list of sockets that are subscribed to the defined
        # channel and removes the current socket from it
        sockets = state.channel_sockets.get(channel, [])
        if socket_id in sockets:
            sockets.remove(socket_id)

        # tries to retrieve the channel data for the channel socket
        # tuple in case there's none available there's nothing else
        # remaining to be done in the unsubscribe process
        channel_data = state.channel_socket_data.get(channel_socket)
        if not channel_data:
            return

        # deletes the channel socket tuple reference from the channel
        # socket data list, no longer going to be required
        del state.channel_socket_data[channel_socket]

        # retrieves both the information on the user id associated with
        # the channel data and the is peer (channel) boolean flag
        user_id = channel_data["user_id"]
        is_peer = channel_data.get("peer", False)

        # gather information on the channel from the global state object
        # this would include the amount of users the members, current
        # connections and more
        info = state.channel_info.get(channel, {})
        users = info.get("users", {})
        members = info.get("members", {})
        conns = info.get("conns", [])
        user_count = info.get("user_count", 0)

        # removes the current connection from the list of connection currently
        # active for the channel, because it's no longer available
        conns.remove(connection)

        # retrieves the currently active connections registered under the user id
        # of the connection to be unregistered then removes the current connection
        # from the list of connections and re-sets the connections list
        user_conns = users.get(user_id, [])
        user_conns.remove(connection)
        users[user_id] = user_conns

        # verifies if the current connection is old, a connection is considered
        # old when no more connections exist for a certain user id in the channel
        # for this situations additional housekeeping must be performed
        is_old = len(user_conns) == 0
        if is_old:
            del users[user_id]
            del members[user_id]
            user_count -= 1

        # updates the various attributes of the channel information structure
        # so that it remains updated according to the unsubscribe operation
        info["users"] = users
        info["members"] = members
        info["conns"] = conns
        info["user_count"] = user_count
        state.channel_info[channel] = info

        # unsubscribes from the complete set of peer channels associated with
        # the current presence channel, this is an expensive operation controlled
        # by the peer flat that may be set in the channel data structure
        is_peer and self.unsubscribe_peer_all(app_key, connection, channel)

        # verifies if the current connection is old in case it's not no operation
        # remain for the unsubscribe operation and so the function may return
        if not is_old:
            return

        # verifies if the current set of connection is empty (count is zero) so that
        # it's possible to know if the channel info for the channel should be removed
        is_empty = len(conns) == 0
        if is_empty:
            del state.channel_info[channel]

        # creates the map that is going to be used in the event to be sent to the set
        # sockets subscribed to the channel indicating that the member has been removed
        json_d = dict(
            event="pusher:member_removed",
            member=json.dumps(channel_data),
            channel=channel,
        )

        # iterates over the complete set of connections subscribed to the channel to notify
        # them about the member that has been removed from the channel
        for _connection in conns:
            # in case the current connection in iteration is the same as the
            # connection in subscription skips the current iteration otherwise
            # send the "member removed" message to the connection
            if _connection == connection:
                continue
            _connection.send_pushi(json_d)

            # in case the current subscription does not have the peer "mode" enabled
            # there's no need to continue as the rest of the iteration is dedicated
            # to the subscription of the peer channels
            if not is_peer:
                continue

            # retrieves the socket id of the current connection in iteration
            # and uses it to construct the channel socket id tuple to try to
            # retrieve the channel data for the socket in case it does not
            # exists skips the current step, no need to subscribe to chat
            # specific channel (because there's no channel data)
            _socket_id = _connection.socket_id
            _channel_socket = (channel, _socket_id)
            _channel_data = state.channel_socket_data.get(_channel_socket)
            if not _channel_data:
                continue

            # retrieves the user id for the connection in iteration for the current
            # channel and uses it to subscribe the connection to the peer channel
            _user_id = _channel_data["user_id"]
            self.unsubscribe_peer(app_key, _connection, channel, user_id, _user_id)

    def validate(self, connection, app_key, socket_id, channel):
        # verifies that the current socket is subscribed of eligible for
        # the provided channel and raises an exception in case it's not,
        # this provides a simple mechanism for security validation
        is_subscribed = self.is_subscribed(app_key, socket_id, channel)
        is_eligible = self.is_eligible(app_key, socket_id, channel)
        if not is_subscribed and not is_eligible:
            raise RuntimeError("Not subscribed/eligible to channel")

    def subscribe_peer_all(self, app_key, connection, channel):
        # creates the channel socket tuple with the channel name and the
        # socket identifier for the current connection
        state = self.get_state(app_key=app_key)
        channel_socket = (channel, connection.socket_id)

        # retrieves the channel data information for the current channel
        # socket and in case there's none returns immediately
        channel_data = state.channel_socket_data.get(channel_socket)
        if not channel_data:
            return

        # retrieves the user identifier from the channel data of the current
        # connection in the channel
        user_id = channel_data["user_id"]

        # uses the channel information to retrieve the list of currently
        # registered connections for the channel, these are going to be
        # used in the subscription iteration
        info = state.channel_info.get(channel, {})
        conns = info.get("conns", [])

        # creates the list that will hold the list of user identifier
        # that have already been visited so that no more that one peer
        # subscription is done by type
        visited = []

        # iterates over all the connections subscribed for the current channel
        # to be able to register for each of the peer channels
        for _connection in conns:
            # in case the current connection in iteration is the connection
            # that is used for the subscription (own connection) skips the
            # current loop as there's nothing to be done
            if _connection == connection:
                continue

            # creates the channel socket tuple containing the channel name
            # and the socket identifier for the current connection in iteration
            # and then uses it to retrieve the channel data for it, in case none
            # is retrieve must skip the current loop
            _channel_socket = (channel, _connection.socket_id)
            _channel_data = state.channel_socket_data.get(_channel_socket)
            if not _channel_data:
                continue

            # retrieves the user identifier for the current channel data in
            # case the user identifier is the same as the current channel's
            # identifiers ignores it (no need to subscribe to our own channel)
            # and then in case it has already been visited also ignores it
            _user_id = _channel_data["user_id"]
            if _user_id == user_id:
                continue
            if _user_id in visited:
                continue

            # subscribes for the peer channel for the user id pair and adds
            # the current user id to the list of visited ids (avoid duplicated
            # subscriptions of channels)
            self.subscribe_peer(app_key, connection, channel, user_id, _user_id)
            visited.append(_user_id)

    def unsubscribe_peer_all(self, app_key, connection, channel):
        # creates the channel socket tuple with the channel name and the
        # socket identifier for the current connection
        state = self.get_state(app_key=app_key)
        channel_socket = (channel, connection.socket_id)

        # retrieves the channel data information for the current channel
        # socket and in case there's none returns immediately
        channel_data = state.channel_socket_data.get(channel_socket)
        if not channel_data:
            return

        # retrieves the user identifier from the channel data of the current
        # connection in the channel
        user_id = channel_data["user_id"]

        # uses the channel information to retrieve the list of currently
        # registered connections for the channel, these are going to be
        # used in the unsubscription iteration
        info = state.channel_info.get(channel, {})
        conns = info.get("conns", [])

        # creates the list that will hold the list of user identifier
        # that have already been visited so that no more that one peer
        # unsubscription is done by type
        visited = []

        # iterates over all the connections subscribed for the current channel
        # to be able to unregister for each of the peer channels
        for _connection in conns:
            # in case the current connection in iteration is the connection
            # that is used for the subscription (own connection) skips the
            # current loop as there's nothing to be done
            if _connection == connection:
                continue

            # creates the channel socket tuple containing the channel name
            # and the socket identifier for the current connection in iteration
            # and then uses it to retrieve the channel data for it, in case none
            # is retrieve must skip the current loop
            _channel_socket = (channel, _connection.socket_id)
            _channel_data = state.channel_socket_data.get(_channel_socket)
            if not _channel_data:
                continue

            # retrieves the user identifier for the current channel data in
            # case the user identifier is the same as the current channel's
            # identifiers ignores it (no need to unsubscribe to our own channel)
            # and then in case it has already been visited also ignores it
            _user_id = _channel_data["user_id"]
            if _user_id == user_id:
                continue
            if _user_id in visited:
                continue

            # unsubscribes from the peer channel for the user id pair and adds
            # the current user id to the list of visited ids (avoid duplicated
            # subscriptions of channels)
            self.unsubscribe_peer(app_key, connection, channel, user_id, _user_id)
            visited.append(_user_id)

    def subscribe_peer(self, app_key, connection, channel, first_id, second_id):
        # in case the id of both parts is the same, must return immediately
        # as this is considered to be an invalid channel, for peer basis
        if first_id == second_id:
            return

        base = [first_id, second_id]
        base.sort()
        base_s = "&".join(base)
        base_channel = channel[9:]
        _channel = "peer-" + base_channel + ":" + base_s
        self.subscribe(connection, app_key, connection.socket_id, _channel, force=True)

    def unsubscribe_peer(self, app_key, connection, channel, first_id, second_id):
        # in case the id of both parts is the same, must return immediately
        # as this is considered to be an invalid channel, for peer basis
        if first_id == second_id:
            return

        base = [first_id, second_id]
        base.sort()
        base_s = "&".join(base)
        base_channel = channel[9:]
        _channel = "peer-" + base_channel + ":" + base_s
        self.unsubscribe(connection, app_key, connection.socket_id, _channel)

    def is_subscribed(self, app_key, socket_id, channel):
        """
        Verifies if the socket identified by the provided socket
        id is subscribed for the provided channel.

        Keep in mind that the channel should be an app id absent
        value and does not identify a channel in an unique way.

        This not a very light operation as it verifies the socket's
        associated channels structure for presence of the channel.
        Use this with care to avoid performance issues.

        :type app_key: String
        :param app_key: The app key to be used in the retrieval of
        the state for the subscription testing.
        :type socket_id: String
        :param socket_id: The identifier of the socket to be checked
        for subscription.
        :type channel: String
        :param channel: The "local" name of the channel to be verified
        for subscription in the current socket context.
        :rtype: bool
        :return: The result of the is subscribed test for the provided
        app key, socket id and channel information.
        """

        state = self.get_state(app_key=app_key)
        channels = state.socket_channels.get(socket_id, None)
        is_subscribed = channel in channels if channels else False
        return is_subscribed

    def is_eligible(self, app_key, socket_id, channel):
        # verifies if the provided channel starts with the peer prefix
        # and if that's not the case returns an invalid value immediately
        # as no eligibility test is possible for other channels
        if not channel.startswith("peer-"):
            return False

        # splits the channel into the base and tail parts and then extracts
        # both the base channel name and the peers list for processing
        base, tail = channel.split(":", 1)
        base_channel = "presence-" + base.split("-", 1)[1]
        peers = tail.split("&", 1)

        # retrieves the required app state from the provided app key, it's
        # going to be used for channel data related operations
        state = self.get_state(app_key=app_key)

        # constructs the channel socket tuple and uses it to retrieve the
        # the channel data value for the base channel so that the user
        # identification is retrieved for verification, note that in no
        # channel data exists the socket is considered not eligible
        channel_socket = (base_channel, socket_id)
        channel_data = state.channel_socket_data.get(channel_socket)
        if not channel_data:
            return False
        user_id = channel_data["user_id"]

        # returns the final boolean value for the presence testing of the
        # user id value in the peers list (as expected)
        return user_id in peers

    def trigger(
        self,
        app_id,
        event,
        data,
        channels=None,
        echo=False,
        persist=False,
        json_d=None,
        owner_id=None,
        verify=True,
    ):
        if not channels:
            channels = ("global",)

        channels_t = type(channels)
        if not channels_t in (list, tuple):
            channels = (channels,)

        invalid = dict()

        for channel in channels:
            self.trigger_c(
                app_id,
                channel,
                event,
                echo,
                persist,
                data,
                json_d=json_d,
                owner_id=owner_id,
                verify=verify,
                invalid=invalid,
            )

    def trigger_c(
        self,
        app_id,
        channel,
        event,
        echo,
        persist,
        data,
        json_d=None,
        owner_id=None,
        verify=True,
        delayed=True,
        invalid={},
    ):
        # retrieves the data type of the provided data of the event and
        # depending on the kind of structure it may dump the result as
        # a plain based string serialized using JSON
        data_t = type(data)
        data = data if data_t in appier.legacy.STRINGS else json.dumps(data)

        # creates the "new" JSON dictionary object that represents the
        # event payload and copies some of the event metadata into it
        # so that it may be consulted latter by the client, note that
        # a copy must be created to avoid any issue related with re-usage
        json_d = dict(json_d) if json_d else dict()
        if channel:
            json_d["channel"] = channel
        if event:
            json_d["event"] = event
        if data:
            json_d["data"] = data

        # in case the persist flag is set the log channel operation is
        # performed so that the event is stored in the data source and
        # may be retrieved latter for reference/observation, note that
        # for these situations the mid and timestamp values are added
        # to the JSON information, so that they may be used for reference
        if persist:
            event_s = self.log_channel(
                app_id,
                channel,
                json_d,
                owner_id=owner_id,
                verify=verify,
                delayed=delayed,
                invalid=invalid,
            )
            json_d["mid"] = event_s.mid
            json_d["timestamp"] = event_s.timestamp

        # runs the process of sending the event through the channel,
        # note that the process will also be run for the complete set
        # of handlers registered for the current infra-structure
        self.send_channel(
            app_id,
            channel,
            json_d,
            echo=echo,
            owner_id=owner_id,
            verify=verify,
            delayed=delayed,
            invalid=invalid,
        )

    def get_subscriptions(self, app_id, channel):
        subscriptions = pushi.Subscription.find(instance=app_id, event=channel)
        return subscriptions

    def log_channel(
        self,
        app_id,
        channel,
        json_d,
        owner_id=None,
        verify=True,
        delayed=True,
        invalid={},
        has_date=True,
    ):
        # verifies that the owner (socket) identifier is present in the channel
        # (but only in case the verify flag is present)
        if owner_id and verify:
            self.verify_presence(app_id, owner_id, channel)

        # generates the proper event structure (includes identifiers
        # and timestamps) to the current event this structure is going
        # to be used for both handling and persistence
        event = self.gen_event(
            app_id, channel, json_d=json_d, owner_id=owner_id, has_date=has_date
        )

        def persist():
            # stores the event structure in the data source for latter usage
            # this is going to be a blocking operation and as such must be
            # used with proper care to avoid blocking of application
            event.save()

            # extracts the mid from the event so that it's does not need
            # to be extracted in every iteration
            mid = event.mid

            # retrieves the complete set of subscriptions for the
            # provided channel and under the current app id to be
            # able to create the proper associations
            subscriptions = self.get_subscriptions(app_id, channel)
            for subscription in subscriptions:
                user_id = subscription.user_id
                if user_id in invalid:
                    continue
                assoc = pushi.Association(instance=app_id, mid=mid, user_id=user_id)
                assoc.save()
                invalid[user_id] = True

        # delays the persistence of the event data and subscription so
        # that the current control flow is not blocked with the data
        # store operations that are going to be performed, in case the
        # delayed flag is not set the operation is executed immediately
        self.app.delay(persist) if delayed else persist()

        # returns the final generated event structure that may be used
        # to retrieve some persistent related information (eg: mid)
        return event

    def send_channel(
        self,
        app_id,
        channel,
        json_d,
        echo=False,
        owner_id=None,
        verify=True,
        delayed=True,
        invalid={},
    ):
        # tries to retrieve the value of the flag that controls if the
        # "custom" handlers should be used for the channel handling
        use_handlers = appier.conf("PUSHI_HANDLERS", True, cast=bool)

        # retrieves the state of the current app to be used in the sending and
        # verifies that the owner (socket) identifier is present in the channel
        # (but only in case the verify flag is present)
        state = self.get_state(app_id=app_id)
        if owner_id and verify:
            self.verify_presence(app_id, owner_id, channel)

        # retrieves the complete set of sockets associated with the channel
        # and sends the JSON data through the socket, avoiding sending the
        # data through the same socket that originated the event, unless the
        # echo flag is set (meaning re-sending to channel)
        sockets = state.channel_sockets.get(channel, [])
        for socket_id in sockets:
            if socket_id in invalid:
                continue
            if socket_id == owner_id and not echo:
                continue
            self.send_socket(socket_id, json_d)
            invalid[socket_id] = True

        # in case the handlers should not be used for custom handling the
        # control flow should be returned immediately as expected
        if not use_handlers:
            return

        # iterates over the complete set of handler currently defined
        # to send the message also through these channels, in case there's
        # a failure the event is logged to avoid unwanted exceptions
        for handler in self.handlers:
            try:
                if delayed:
                    self.app.delay(
                        handler.send,
                        args=(app_id, channel, json_d),
                        kwargs=dict(invalid=invalid),
                    )
                else:
                    handler.send(app_id, channel, json_d, invalid=invalid)
            except Exception as exception:
                self.app.logger.info(
                    "Problem using handler '%s' for sending - %s"
                    % (handler.name, appier.legacy.UNICODE(exception))
                )

    def send_socket(self, socket_id, json_d):
        self.server.send_socket(socket_id, json_d)

    def get_state(self, app_id=None, app_key=None):
        # sets the initial value of the state value as invalid so
        # that by default the returned value is invalid
        state = None

        # in case no app id and no app key are provided there's no
        # way to retrieve the app state (no identifier) and as such
        # must raise an exception indicating the problem
        if not app_id and not app_key:
            raise RuntimeError("No app identifier was provided")

        # retrieves the state object taking into account first the
        # app id and then the app key (as an alternative) after this
        # call the state object should be populated
        if app_id:
            state = self.app_id_state.get(app_id, None)
        elif app_key:
            state = self.app_key_state.get(app_key, None)

        # in case the state object is found (already created) returns
        # the value immediately (to the caller method)
        if state:
            return state

        # retrieves the app object for the provided parameters and in case
        # the app is not found raises an exception indicating so
        app = self.get_app(app_id=app_id, app_key=app_key)
        if not app:
            raise RuntimeError("No app found for the provided parameters")

        # unpacks both the app id and key values from the app object
        # to be used in the construction of the app state object
        app_id = app.ident
        app_key = app.key

        # creates the (app) state object with the provided app id and key
        # and the updates the associated dictionaries to access the app state
        # from both the app id and key values
        state = AppState(app_id, app_key)
        self.app_id_state[app_id] = state
        self.app_key_state[app_key] = state

        # returns the creates state object to the caller method, as requested
        # (this state object has just been created)
        return state

    def get_channel(self, app_key, channel, skip=0, count=10, limit=True):
        members = self.get_members(app_key, channel)
        alias = self.get_alias(app_key, channel)
        events = self.get_events(app_key, channel, skip=skip, count=count, limit=limit)
        return dict(name=channel, members=members, alias=alias, events=events)

    def get_members(self, app_key, channel):
        state = self.get_state(app_key=app_key)
        info = state.channel_info.get(channel, {})
        members = info.get("members", {})
        return members

    def get_alias(self, app_key, channel):
        state = self.get_state(app_key=app_key)
        return state.alias.get(channel, [])

    def get_channels(self, app_key, alias):
        state = self.get_state(app_key=app_key)
        return state.alias_i.get(alias, [])

    def get_events(self, app_key, channel, skip=0, count=10, limit=True, map=True):
        is_personal = channel.startswith("personal-")
        if not is_personal and limit:
            return []

        if is_personal:
            return self.get_events_personal(
                app_key, channel, skip=skip, count=count, map=map
            )
        else:
            return self.get_events_global(
                app_key, channel, skip=skip, count=count, map=map
            )

    def get_events_global(self, app_key, channel, skip=0, count=10, map=True):
        app_id = self.app_key_to_app_id(app_key)
        events = pushi.PushiEvent.find(
            instance=app_id,
            channel=channel,
            skip=skip,
            limit=count,
            sort=[("id", -1)],
            map=map,
        )
        for event in events:
            del event["_id"]
        return events

    def get_events_personal(self, app_key, channel, skip=0, count=10, map=True):
        user_id = channel[9:]
        app_id = self.app_key_to_app_id(app_key)

        assocs = pushi.Association.find(
            instance=app_id, user_id=user_id, skip=skip, limit=count, sort=[("id", -1)]
        )
        mids = [assoc.mid for assoc in assocs]

        events = pushi.PushiEvent.find(
            instance=app_id, mid={"$in": mids}, sort=[("id", -1)], map=map
        )
        for event in events:
            del event["_id"]

        return events

    def get_app(self, app_id=None, app_key=None, raise_e=True):
        if app_id:
            app = pushi.App.get(ident=app_id, raise_e=raise_e)
        if app_key:
            app = pushi.App.get(key=app_key, raise_e=raise_e)
        return app

    def has_app(self, app_id=None, app_key=None):
        # verifies if state is currently defined for the
        # provided app values and if that's the case returns
        # immediately with a valid value
        if self.has_state(app_id=app_id, app_key=app_key):
            return True

        # tries to retrieve an app for the provided values
        # and in case there's one an app is considered to
        # exist (returns a valid value)
        app = self.get_app(app_id=app_id, app_key=app_key, raise_e=False)
        if app:
            return True

        # by default an invalid value is returned as none of the
        # verifications is considered successful
        return False

    def has_state(self, app_id=None, app_key=None):
        # verifies if either the app id or the app key are
        # present in the respective state maps and if that's
        # the case returns a valid value (presence validated)
        if app_id in self.app_id_state:
            return True
        if app_key in self.app_key_state:
            return True
        return False

    def invalidate(self, app_id=None, app_key=None):
        """
        Invalidates the state of the app for the provided identifier
        and key values. Note that the invalidated partials are only
        affected when the provided identification and key parts are
        properly provided.

        This method should be used with extreme care as unwanted results
        may emerge from wreckless usage of this method.

        :type app_id: String
        :param app_id: The identifier of the application that is going
        to be invalidated for the current state.
        :type app_key: String
        :param app_key: The key of the application that is going
        to be invalidated for the current state.
        """

        if app_id and app_id in self.app_id_state:
            del self.app_id_state[app_id]
        if app_key and app_key in self.app_id_state:
            del self.app_id_state[app_key]

    def verify(self, app_key, socket_id, channel, auth):
        """
        Verifies the provided auth (token) using the app
        secret associated with the app with the provided
        app key.

        This operation is required for the private channels
        so that only the authenticated user are allowed.

        The verification operation will raise an exception in
        the signature generated is not valid (verification has
        failed for security reasons).

        :type app_key: String
        :param app_key: The app key for the app that is going
        to be used as the base for the verification.
        :type socket_id: String
        :param socket_id: The identifier of the socket that is
        going to be used in the process of verification.
        :type channel: String
        :param channel: The name of the channel that is going
        to be used in the verification process.
        :type auth: String
        :param auth: The string that is going to be used for auth
        this should be an HMAC based token string.
        """

        app = self.get_app(app_key=app_key)
        app_secret = app.secret
        app_secret = appier.legacy.bytes(str(app_secret))

        string = "%s:%s" % (socket_id, channel)
        string = appier.legacy.bytes(string)
        structure = hmac.new(app_secret, string, hashlib.sha256)
        digest = structure.hexdigest()
        auth_v = "%s:%s" % (app_key, digest)

        if not auth == auth_v:
            raise RuntimeError("Invalid signature")

    def verify_presence(self, app_id, socket_id, channel):
        state = self.get_state(app_id=app_id)
        channels = state.socket_channels.get(socket_id, [])
        if not channel in channels:
            raise RuntimeError(
                "Socket '%s' is not allowed for '%s'" % (socket_id, channel)
            )

    def app_id_to_app_key(self, app_id):
        state = self.get_state(app_id=app_id)
        return state.app_key

    def app_key_to_app_id(self, app_key):
        state = self.get_state(app_key=app_key)
        return state.app_id

    def gen_event(self, app_id, channel, json_d, owner_id=None, has_date=True):
        """
        Generates the complete event structure from the provided
        details on the current context.

        Anyone using this method should not expect the same
        results from two different calls as this method includes
        some random string generation.

        :type app_id: String
        :param app_id: The identifier of the app that is currently
        being used for the the event sending.
        :type channel: String
        :param channel: The name of the channel that is going to be
        used for sending the event.
        :type json_d: Dictionary
        :param json_d: The map containing all the (payload) information
        that is the proper event.
        :type owner_id: String
        :param owner_id: The identifier used by the entity that "owns"
        the event to be sent.
        :type has_date: bool
        :param has_date: If the generates event structure should include
        the data in its structure, this account for more processing.
        :rtype: PushiEvent
        :return: The generated event structure that was created according
        to the provided details for generation.
        """

        # runs a series of initial pre condition verification to ensure that
        # the data is ready to be persisted if required
        appier.verify(not "mid" in json_d)
        appier.verify(not "timestamp" in json_d)

        # generates a globally unique identifier that is going to be the
        # sole unique value for the event, this may be used latter for
        # unique unique identification
        mid = str(uuid.uuid4())

        # generates a timestamp that is going to identify the timing of the
        # event this value should not be trusted as this does not represent
        # the time of sending of the event but instead the generation of
        # the global event structure
        timestamp = time.time()

        # creates the proper dictionary of the event that includes all of the
        # main values of it, together with the ones that have been generated
        event = pushi.PushiEvent(
            mid=mid,
            instance=app_id,
            channel=channel,
            owner_id=owner_id,
            timestamp=timestamp,
            data=dict(json_d),
        )

        # in case the date inclusion flag is set a new date string must be
        # generates and attached to the current event structure (extra values)
        if has_date:
            date = datetime.datetime.utcfromtimestamp(timestamp)
            date_s = date.strftime("%B %d, %Y %H:%M:%S UTC")
            event.date = date_s

        # returns the "final" event structure to the caller method so that it
        # can be used as the complete event structure
        return event


if __name__ == "__main__":
    state = State()
    app = pushi.PushiApp(state)
    server = pushi.PushiServer(state)
    state.load(app, server)
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """


class APNAPI(object):
    def create_apn(self, token, event, auth=None, unsubscribe=True):
        # runs the APN subscription operation for the provided
        # token and event, this operation uses the currently
        # defined app id for the operation, then returns the
        # resulting dictionary to the caller method
        result = self.post(
            self.base_url + "apns",
            params=dict(auth=auth, unsubscribe=unsubscribe),
            data_j=dict(token=token, event=event),
        )
        return result

    def delete_apn(self, token, event):
        # runs the unsubscription operation for the provided
        # user id and event, this operation uses the currently
        # defined app id for the operation, then returns the
        # resulting dictionary to the caller method
        result = self.delete(self.base_url + "apns/%s/%s" % (token, event))
        return result

    def subscribe_apn(self, *args, **kwargs):
        return self.create_apn(*args, **kwargs)

    def unsubscribe_apn(self, *args, **kwargs):
        return self.delete_apn(*args, **kwargs)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier

import pushi


class APNController(appier.Controller):
    @appier.private
    @appier.route("/apns", "GET")
    def list(self):
        token = self.field("token", None)
        event = self.field("event", None)
        return self.state.apn_handler.subscriptions(token=token, event=event)

    @appier.private
    @appier.route("/apns", "POST")
    def create(self):
        auth = self.field("auth", None)
        unsubscribe = self.field("unsubscribe", False, cast=bool)
        apn = pushi.APN.new()
        apn = self.state.apn_handler.subscribe(apn, auth=auth, unsubscribe=unsubscribe)
        return apn.map()

    @appier.private
    @appier.route("/apns/<token>", "DELETE")
    def deletes(self, token):
        apns = self.state.apn_handler.unsubscribes(token)
        return dict(subscriptions=[apn.map() for apn in apns])

    @appier.private
    @appier.route("/apns/<token>/<regex('[\.\w-]+'):event>", "DELETE")
    def delete(self, token, event):
        force = self.field("force", False, cast=bool)
        apn = self.state.apn_handler.unsubscribe(token, event=event, force=force)
        return apn.map()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier

import pushi


class AppController(appier.Controller):
    @appier.private
    @appier.route("/apps", "GET")
    def list(self):
        apps = pushi.App.find(map=True)
        return dict(apps=apps)

    @appier.route("/apps", "POST")
    def create(self):
        app = pushi.App.new()
        app.save()
        return app.map()

    @appier.private
    @appier.route("/apps/<ident>", "GET")
    def show(self, ident):
        app = pushi.App.get(map=True, ident=ident)
        return app

    @appier.private
    @appier.route("/apps/<ident>", "PUT")
    def update(self, ident):
        app = pushi.App.get(ident=ident)
        app.apply()
        app.save()
        return app

    @appier.private
    @appier.route("/apps/ping", "GET")
    def ping(self):
        app = pushi.App.get()
        self.state.trigger(app.id, "ping", "ping", persist=False)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """


class AppAPI(object):
    def create_app(self, name):
        # runs the post call that will create the app with the provided
        # name then returns the returning map to the caller method, it
        # should contain the generated information for the app
        result = self.post(self.base_url + "apps", auth=False, data_j=dict(name=name))
        return result

    def update_app(self, app_id=None, **kwargs):
        # retrieves the proper app id to be used defaulting to the current
        # defined app id in case none is provided
        app_id = app_id or self.app_id

        # runs the pit call that will create the app with the provided
        # name then returns the returning map to the caller method, it
        # should contain the newly updated information for the app
        result = self.put(self.base_url + "apps/%s" % app_id, data_j=kwargs)
        return result



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import uuid
import hashlib

import appier

from . import base


class App(base.PushiBase):
    name = appier.field(index=True, default=True)

    ident = appier.field(index=True, safe=True, immutable=True)

    key = appier.field(index=True, safe=True, immutable=True)

    secret = appier.field(index=True, safe=True, immutable=True)

    apn_sandbox = appier.field(
        type=bool,
        description="APN Sandbox",
        observations="""Indicates if the APN context is sandbox or
        production related, should be in sync with Apple Development configuration""",
    )

    apn_key = appier.field(
        meta="longtext",
        description="APN Key",
        observations="""The private key in PEM format to be used
        in messages to be sent using APN (Apple Push Notifications)""",
    )

    apn_cer = appier.field(
        meta="longtext",
        description="APN Cer",
        observations="""The certificate in PEM format to be used
        in messages to be sent using APN (Apple Push Notifications)""",
    )

    @classmethod
    def validate(cls):
        return super(App, cls).validate() + [
            appier.not_null("name"),
            appier.not_empty("name"),
            appier.not_duplicate("name", cls._name()),
        ]

    @classmethod
    def list_names(cls):
        return ["name", "ident", "apn_sandbox"]

    def pre_create(self):
        base.PushiBase.pre_create(self)

        ident = appier.legacy.bytes(str(uuid.uuid4()))
        key = appier.legacy.bytes(str((uuid.uuid4())))
        secret = appier.legacy.bytes(str(uuid.uuid4()))

        self.ident = hashlib.sha256(ident).hexdigest()
        self.key = hashlib.sha256(key).hexdigest()
        self.secret = hashlib.sha256(secret).hexdigest()

        self.instance = self.ident



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import hmac
import hashlib

import appier

from . import apn
from . import app
from . import web
from . import event
from . import subscription

BASE_URL = "https://puxiapp.com:9090/"
""" The base URL to be used by the API to access
the remote endpoints, should not be changed """

BASE_WS_URL = "wss://puxiapp.com/"
""" The default base websockets URL that is going
to be used in case no other value is specified """


class API(
    appier.API,
    apn.APNAPI,
    app.AppAPI,
    web.WebAPI,
    event.EventAPI,
    subscription.SubscriptionAPI,
):
    """
    Base class for the construction of the pushi
    proxy object for interaction with the server
    side of the pushi system.

    Should provide the various methods that enable
    the developer to make operations for both read
    and write (create and read).
    """

    def __init__(self, *args, **kwargs):
        appier.API.__init__(self, *args, **kwargs)
        self.app_id = appier.conf("PUSHI_ID", None)
        self.app_key = appier.conf("PUSHI_KEY", None)
        self.app_secret = appier.conf("PUSHI_SECRET", None)
        self.base_url = appier.conf("PUSHI_URL", BASE_URL)
        self.app_id = kwargs.get("app_id", self.app_id)
        self.app_key = kwargs.get("app_key", self.app_key)
        self.app_secret = kwargs.get("app_secret", self.app_secret)
        self.base_url = kwargs.get("base_url", self.base_url)
        self.token = None

    def build(
        self,
        method,
        url,
        data=None,
        data_j=None,
        data_m=None,
        headers=None,
        params=None,
        mime=None,
        kwargs=None,
    ):
        auth = kwargs.pop("auth", True)
        if auth:
            kwargs["sid"] = self.get_token()

    def get_token(self):
        if self.token:
            return self.token
        return self.login()

    def auth_callback(self, params, headers):
        token = self.login()
        params["sid"] = token

    def login(self):
        # tries to login in the pushi infra-structure using the
        # login route together with the full set of auth info
        # retrieving the result map that should contain the
        # session token, to be used in further calls
        result = self.get(
            self.base_url + "login",
            callback=False,
            auth=False,
            app_id=self.app_id,
            app_key=self.app_key,
            app_secret=self.app_secret,
        )

        # unpacks the token value from the result map and then
        # returns the token to the caller method
        self.token = result["token"]
        return self.token

    def logout(self):
        # runs the "simplistic" call to the logout operation so
        # that the session is invalidated from the server side
        self.get(self.base_url + "logout")

        # invalidates the currently set token so that it's no longer
        # going to be used for any kind of operation
        self.token = None

    def authenticate(self, channel, socket_id):
        # in case the app key is not defined for the current
        # instance an exception must be raised as it's not possible
        # to run the authentication process without an app key
        if not self.app_key:
            raise RuntimeError("No app key defined")

        # creates the string to hashed using both the provided
        # socket id and channel (concatenation)
        string = "%s:%s" % (socket_id, channel)
        string = appier.legacy.bytes(string)

        # runs the HMAC encryption in the provided secret and
        # the constructed string and returns a string containing
        # both the key and the hexadecimal digest
        app_secret = appier.legacy.bytes(str(self.app_secret))
        structure = hmac.new(app_secret, string, hashlib.sha256)
        digest = structure.hexdigest()
        return "%s:%s" % (self.app_key, digest)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier

import pushi


class SubscriptionController(appier.Controller):
    @appier.private
    @appier.route("/subscriptions", "GET")
    def list(self):
        user_id = self.field("user_id", None)
        event = self.field("event", None)
        filter = dict()
        if user_id:
            filter["user_id"] = user_id
        if event:
            filter["event"] = event
        subscriptions = pushi.Subscription.find(map=True, **filter)
        return dict(subscriptions=subscriptions)

    @appier.private
    @appier.route("/subscriptions", "POST")
    def create(self):
        subscription = pushi.Subscription.new()
        exists = pushi.Subscription.exists(
            user_id=subscription.user_id, event=subscription.event
        )
        if exists:
            subscription = exists
        else:
            subscription.save()
        return subscription.map()

    @appier.private
    @appier.route("/subscriptions/<user_id>/<regex('[\.\w-]+'):event>", "DELETE")
    def delete(self, user_id, event):
        force = self.field("force", False, cast=bool)
        subscription = pushi.Subscription.get(
            user_id=user_id, event=event, raise_e=force
        )
        if not subscription:
            return None
        subscription.delete()
        return subscription.map()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier

import pushi


class EventController(appier.Controller):
    @appier.private
    @appier.route("/events", "GET")
    def list(self):
        count = self.field("count", 10, cast=int)
        events = pushi.PushiEvent.find(
            limit=count,
            sort=[
                ("id", -1),
            ],
            map=True,
        )
        return dict(events=events)

    @appier.private
    @appier.route("/events", "POST")
    def create(self):
        data = self.request.get_json() or dict()
        app_id = self.session.get("app_id", None)
        _data = data.get("data", None)
        event = data.get("event", "message")
        channel = data.get("channel", "global")
        persist = data.get("persist", True)
        if not _data:
            raise RuntimeError("No data set for event")
        self.state.trigger(
            app_id,
            event,
            _data,
            channels=channel,
            persist=persist,
            json_d=data,
            verify=False,
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import shutil
import tempfile

import netius.clients

import pushi

from . import handler


class APNHandler(handler.Handler):
    """
    Pushi handler (adapter) for the Apple Push Notifications
    (APN) infra-structure, so that it's possible to send
    push notification to iOS/OSX devices.
    """

    def __init__(self, owner):
        handler.Handler.__init__(self, owner, name="apn")
        self.subs = {}

    def send(self, app_id, event, json_d, invalid={}):
        # tries to retrieve the appropriate message starting from
        # the most general values to the most specific values,
        # be aware that the value may be a bit fuzzy
        message = json_d.get("data", None)
        message = json_d.get("push", message)
        message = json_d.get("apn", message)
        if not message:
            raise RuntimeError("No message defined")

        # retrieves the reference to the app with the defined app id
        # and extracts the APN specific values for it to be used in
        # the process of authentication
        app = self.owner.get_app(app_id=app_id)
        key_data = app.apn_key
        cer_data = app.apn_cer
        sandbox = app.apn_sandbox

        # in case no key data or certificate data is present a runtime
        # error is raised to indicate the problem
        if not key_data:
            raise RuntimeError("No APN key defined")
        if not cer_data:
            raise RuntimeError("No APN certificate defined")

        # ensures that the complete set of data is encoded as bytes, as
        # this is required for the proper writing of the file, otherwise
        # and invalid encoding would be raised for some platforms
        key_data = netius.legacy.bytes(key_data)
        cer_data = netius.legacy.bytes(cer_data)

        # creates a new temporary directory that will be used to store
        # the temporary key and certificate files for SSL
        path = tempfile.mkdtemp()

        # creates the full paths to both the key and certificate
        # files using the temporary path as base
        key_path = os.path.join(path, "apn.key")
        cer_path = os.path.join(path, "apn.cer")

        # opens the SSL key file for writing (in binary mode) and
        # then writes the current data into it so that it may be
        # used by the encryption infra-structure
        key_file = open(key_path, "wb")
        try:
            key_file.write(key_data)
        finally:
            key_file.close()

        # opens the temporary certificate file and writes
        # the retrieved certificate data into it, to be used
        # temporarily by the SSL infra-structure
        cer_file = open(cer_path, "wb")
        try:
            cer_file.write(cer_data)
        finally:
            cer_file.close()

        # retrieves the app key for the retrieved app by unpacking the current
        # app structure into the appropriate values
        app_key = app.key

        # saves the original event name for the received event, so that it may
        # be used latter for debugging/log purposes
        root_event = event

        # resolves the complete set of (extra) channels for the provided
        # event assuming that it may be associated with alias, then creates
        # the complete list of event containing also the "extra" events
        extra = self.owner.get_channels(app_key, event)
        events = [event] + extra

        # retrieves the complete set of subscriptions for the current APN
        # infra-structure to be able to resolve the appropriate tokens
        subs = self.subs.get(app_id, {})

        # creates the initial list of tokens to be notified and then populates
        # the list with the various token associated with the complete set of
        # resolved events, note that a set is created at the end so that one
        # token gets notified only once (no double notifications)
        tokens = []
        for event in events:
            _tokens = subs.get(event, [])
            tokens.extend(_tokens)
        tokens = set(tokens)
        count = len(tokens)

        # prints a logging message about the various (web) subscriptions
        # that were found for the event that was triggered
        self.logger.debug("Found %d APN subscription(s) for '%s'" % (count, root_event))

        # creates the counter that will be used by the cleanup function
        # to know exactly when to remove the SSL associated files
        pending = len(tokens)
        clojure = dict(pending=pending)

        # creates the cleanup function that will be called for
        # the close operation of the APN protocol, this function
        # will remove the temporary path when called as a response
        # to the "stop" of the last protocol
        def cleanup(protocol):
            # schedules the stop of the current event loop as the
            # APN operation has been completed with success
            netius.compat_loop(loop).stop()

            # retrieve the number of pending APN messages from the
            # clojure and verifies if the end has been reached
            pending = clojure["pending"]
            pending -= 1
            clojure["pending"] = pending
            if not pending == 0:
                return

            # removes the temporary keys and certificated files as
            # they are no longer required for APN usage
            self.logger.debug("Removing temporary APN keys")
            shutil.rmtree(path, ignore_errors=True)

        # iterates over the complete set of tokens to be notified and notifies
        # them using the current APN client infra-structure
        for token in tokens:
            # in case the current token is present in the current
            # map of invalid items must skip iteration as the message
            # has probably already been sent "to the token"
            if token in invalid:
                continue

            # prints a debug message about the APN message that
            # is going to be sent (includes token)
            self.logger.debug("Sending APN message to '%s'" % token)

            # creates the new APN protocol to be used and uses it to
            # send the new message (should be correctly serialized)
            loop, protocol = netius.clients.APNClient.notify_s(
                token,
                message=message,
                sandbox=sandbox,
                key_file=key_path,
                cer_file=cer_path,
            )
            protocol.bind("finish", cleanup)
            loop.run_forever()

            # adds the current token to the list of invalid item for
            # the current message sending stream
            invalid[token] = True

    def load(self):
        subs = pushi.APN.find()
        for sub in subs:
            app_id = sub.app_id
            token = sub.token
            event = sub.event
            self.add(app_id, token, event)

    def add(self, app_id, token, event):
        events = self.subs.get(app_id, {})
        tokens = events.get(event, [])
        tokens.append(token)
        events[event] = tokens
        self.subs[app_id] = events

    def remove(self, app_id, token, event):
        events = self.subs.get(app_id, {})
        tokens = events.get(event, [])
        if token in tokens:
            tokens.remove(token)

    def subscriptions(self, token=None, event=None):
        filter = dict()
        if token:
            filter["token"] = token
        if event:
            filter["event"] = event
        subscriptions = pushi.APN.find(map=True, **filter)
        return dict(subscriptions=subscriptions)

    def subscribe(self, apn, auth=None, unsubscribe=True):
        self.logger.debug("Subscribing '%s' for '%s'" % (apn.token, apn.event))

        is_private = (
            apn.event.startswith("private-")
            or apn.event.startswith("presence-")
            or apn.event.startswith("peer-")
            or apn.event.startswith("personal-")
        )

        is_private and self.owner.verify(apn.app_key, apn.token, apn.event, auth)
        unsubscribe and self.unsubscribe(apn.token, force=False)

        exists = pushi.APN.exists(token=apn.token, event=apn.event)
        if exists:
            apn = exists
        else:
            apn.save()

        self.logger.debug("Subscribed '%s' for '%s'" % (apn.token, apn.event))

        return apn

    def unsubscribe(self, token, event=None, force=True):
        self.logger.debug("Unsubscribing '%s' from '%s'" % (token, event or "*"))

        kwargs = dict(token=token, raise_e=force)
        if event:
            kwargs["event"] = event

        apn = pushi.APN.get(**kwargs)
        if not apn:
            return None

        apn.delete()

        self.logger.debug("Unsubscribed '%s' for '%s'" % (token, event or "*"))

        return apn

    def unsubscribes(self, token, event=None):
        kwargs = dict(token=token)
        if event:
            kwargs["event"] = event

        apns = pushi.APN.find(**kwargs)
        for apn in apns:
            apn.delete()

        return apns



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """


class EventAPI(object):
    def create_event(self, channel, data, event="message", persist=True, **kwargs):
        # creates the initial JSON data structure to be used as the message
        # and then "extends" it with the extra key word arguments passed
        # to this methods as a method of extension
        data_j = dict(data=data, event=event, channel=channel, persist=persist)
        for key in kwargs:
            data_j[key] = kwargs[key]

        # performs the concrete event trigger operation creating an event
        # with the provided information using a secure channel
        result = self.post(self.base_url + "events", data_j=data_j)
        return result

    def trigger_event(self, *args, **kwargs):
        return self.create_event(*args, **kwargs)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import appier

from . import base


class PushiEvent(base.PushiBase):
    mid = appier.field(index=True, immutable=True, default=True, description="MID")

    channel = appier.field(index=True, immutable=True)

    owner_id = appier.field(immutable=True, description="Owner ID")

    timestamp = appier.field(type=float, index=True, immutable=True, meta="datetime")

    data = appier.field(type=dict, immutable=True, meta="longtext")

    @classmethod
    def validate(cls):
        return super(PushiEvent, cls).validate() + [
            appier.not_null("mid"),
            appier.not_empty("mid"),
            appier.not_null("channel"),
            appier.not_empty("channel"),
            appier.not_null("timestamp"),
        ]

    @classmethod
    def list_names(cls):
        return ["mid", "channel", "owner_id", "timestamp"]

    def pre_save(self):
        base.PushiBase.pre_save(self)
        appier.verify(not "mid" in self.data)
        appier.verify(not "timestamp" in self.data)

    @classmethod
    @appier.operation(
        name="Trigger",
        description="""Triggers a new event on the pushi system running
        the complete set of handlers associated""",
        parameters=(
            ("App ID", "app_id", str),
            ("Event", "event", str),
            ("Data", "data", "longtext"),
            ("Persist", "persist", bool, True),
        ),
    )
    def trigger_s(cls, app_id, event, data, persist=True):
        state = appier.get_app().state
        state.trigger(app_id, event, data, persist=persist)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Pushi System
# Copyright (c) 2008-2024 Hive Solutions Lda.
#
# This file is part of Hive Pushi System.
#
# Hive Pushi System is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Pushi System is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Pushi System. If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__copyright__ = "Copyright (c) 2008-2024 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import uuid
import json

import netius.servers


class PushiConnection(netius.servers.WSConnection):
    def __init__(self, *args, **kwargs):
        netius.servers.WSConnection.__init__(self, *args, **kwargs)
        self.app_key = None
        self.socket_id = str(uuid.uuid4())
        self.channels = []
        self.count = 0

    def send_pushi(self, json_d):
        data = json.dumps(json_d)
        self.send_ws(data)
        self.count += 1
        self.owner.count += 1

    def load_app(self):
        app_key = self.path.rsplit("/", 1)[-1]
        is_unicode = netius.legacy.is_unicode(app_key)
        if not is_unicode:
            app_key = app_key.decode("utf-8")
        if not app_key:
            raise RuntimeError("Invalid app key loaded")
        if not len(app_key) == 64:
            raise RuntimeError("Invalid app key length")
        self.app_key = app_key


class PushiServer(netius.servers.WSServer):
    WS_CLOSE_FRAME = b"\x03\xe9"

    def __init__(self, state=None, *args, **kwargs):
        netius.servers.WSServer.__init__(self, *args, **kwargs)
        self.state = state
        self.sockets = {}
        self.count = 0

    def info_dict(self):
        info = netius.servers.WSServer.info_dict(self)
        info["count"] = self.count
        return info

    def on_connection_c(self, connection):
        netius.servers.WSServer.on_connection_c(self, connection)
        self.sockets[connection.socket_id] = connection
        self.trigger(
            "connect",
            connection=connection,
            app_key=connection.app_key,
            socket_id=connection.socket_id,
        )

    def on_connection_d(self, connection):
        netius.servers.WSServer.on_connection_d(self, connection)
        del self.sockets[connection.socket_id]
        self.trigger(
            "disconnect",
            connection=connection,
            app_key=connection.app_key,
            socket_id=connection.socket_id,
        )

    def build_connection(self, socket, address, ssl=False):
        return PushiConnection(self, socket, address, ssl=ssl)

    def on_handshake(self, connection):
        netius.servers.WSServer.on_handshake(self, connection)
        connection.load_app()

        json_d = dict(
            event="pusher:connection_established",
            data=json.dumps(dict(socket_id=connection.socket_id)),
        )
        connection.send_pushi(json_d)

    def on_data_ws(self, connection, data):
        netius.servers.WSServer.on_data_ws(self, connection, data)

        cls = self.__class__

        if not data:
            return
        if data == cls.WS_CLOSE_FRAME:
            return

        try:
            data = data.decode("utf-8")
            json_d = json.loads(data)
        except Exception:
            raise netius.DataError("Invalid message received '%s'" % data)

        event = json_d.get("event", None)
        event = event.replace(":", "_")

        method_name = "handle_" + event
        has_method = hasattr(self, method_name)
        if has_method:
            method = getattr(self, method_name)
        else:
            method = self.handle_event
        method(connection, json_d)

    def handle_pusher_subscribe(self, connection, json_d):
        data = json_d.get("data", {})
        channel = data.get("channel", None)
        auth = data.get("auth", None)
        channel_data = data.get("channel_data", None)

        self.trigger(
            "subscribe",
            connection=connection,
            app_key=connection.app_key,
            socket_id=connection.socket_id,
            channel=channel,
            auth=auth,
            channel_data=channel_data,
        )

        if not self.state:
            return

        data = self.state.get_channel(connection.app_key, channel)
        json_d = dict(
            event="pusher_internal:subscription_succeeded",
            data=json.dumps(data),
            channel=channel,
        )
        connection.send_pushi(json_d)

    def handle_pusher_unsubscribe(self, connection, json_d):
        data = json_d.get("data", {})
        channel = data.get("channel", None)

        self.trigger(
            "unsubscribe",
            connection=connection,
            app_key=connection.app_key,
            socket_id=connection.socket_id,
            channel=channel,
        )

        if not self.state:
            return

        data = self.state.get_channel(connection.app_key, channel)
        json_d = dict(
            event="pusher_internal:unsubscription_succeeded",
            data=json.dumps(data),
            channel=channel,
        )
        connection.send_pushi(json_d)

    def handle_pusher_latest(self, connection, json_d):
        data = json_d.get("data", {})
        channel = data.get("channel", None)
        skip = data.get("skip", 0)
        count = data.get("count", 10)

        self.trigger(
            "validate",
            connection=connection,
            app_key=connection.app_key,
            socket_id=connection.socket_id,
            channel=channel,
        )

        if not self.state:
            return

        data = self.state.get_channel(
            connection.app_key, channel, skip=skip, count=count, limit=False
        )
        json_d = dict(
            event="pusher_internal:latest", data=json.dumps(data), channel=channel
        )
        connection.send_pushi(json_d)

    def handle_event(self, connection, json_d):
        data = json_d["data"]
        event = json_d["event"]
        channel = json_d["channel"]
        echo = json_d.get("echo", False)
        persist = json_d.get("persist", True)

        if not self.state:
            return

        app_id = self.state.app_key_to_app_id(connection.app_key)
        self.state.trigger(
            app_id,
            event,
            data,
            channels=(channel,),
            echo=echo,
            persist=persist,
            owner_id=connection.socket_id,
        )

    def send_socket(self, socket_id, json_d):
        connection = self.sockets[socket_id]
        connection.send_pushi(json_d)


if __name__ == "__main__":
    server = PushiServer()
    server.serve()
else:
    __path__ = []
